import argparse
import json
import logging
import sys
import uuid
import zipfile
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import annofabapi
from annofabapi.dataclass.annotation import AdditionalData, AnnotationDetail, FullAnnotationData
from annofabapi.models import (
    AdditionalDataDefinitionType,
    AdditionalDataDefinitionV1,
    AnnotationDataHoldingType,
    LabelV1,
    ProjectMemberRole,
    TaskStatus,
)
from annofabapi.parser import (
    SimpleAnnotationParser,
    SimpleAnnotationParserByTask,
    lazy_parse_simple_annotation_dir_by_task,
    lazy_parse_simple_annotation_zip_by_task,
)
from annofabapi.utils import can_put_annotation
from dataclasses_json import DataClassJsonMixin
from more_itertools import first_true

import annofabcli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import AbstractCommandLineInterface, ArgumentParser, build_annofabapi_resource_and_login
from annofabcli.common.visualize import AddProps, MessageLocale

logger = logging.getLogger(__name__)


@dataclass
class ImportedSimpleAnnotationDetail(DataClassJsonMixin):
    """
    ``annofabapi.dataclass.annotation.SimpleAnnotationDetail`` に対応するインポート用のDataClass。
    """

    label: str
    """アノテーション仕様のラベル名(英語)"""

    annotation_id: Optional[str]
    """アノテーションID"""

    data: Dict[str, Any]
    """"""

    attributes: Dict[str, Union[str, bool, int]]
    """属性情報。キーは属性の名前、値は属性の値。 """


@dataclass
class ImportedSimpleAnnotation(DataClassJsonMixin):
    """
    ``annofabapi.dataclass.annotation.SimpleAnnotation`` に対応するインポート用のDataClass。
    """

    details: List[ImportedSimpleAnnotationDetail]
    """矩形、ポリゴン、全体アノテーションなど個々のアノテーションの配列。"""


class ImportAnnotation(AbstractCommandLineInterface):
    """
    アノテーションをインポートする
    """

    def __init__(self, service: annofabapi.Resource, facade: AnnofabApiFacade, args: argparse.Namespace):
        super().__init__(service, facade, args)
        self.visualize = AddProps(self.service, args.project_id)

    def get_label_info_from_label_name(self, label_name: str) -> Optional[LabelV1]:
        for label in self.visualize.specs_labels:
            label_name_en = self.visualize.get_label_name(label["label_id"], MessageLocale.EN)
            if label_name_en is not None and label_name_en == label_name:
                return label

        logger.warning(f"アノテーション仕様に label_name={label_name} のラベルが存在しません。")
        return None

    def _get_additional_data_from_attribute_name(
        self, attribute_name: str, label_info: LabelV1
    ) -> Optional[AdditionalDataDefinitionV1]:
        for additional_data in label_info["additional_data_definitions"]:
            additional_data_name_en = self.visualize.get_additional_data_name(
                additional_data["additional_data_definition_id"], MessageLocale.EN, label_id=label_info["label_id"]
            )
            if additional_data_name_en is not None and additional_data_name_en == attribute_name:
                return additional_data

        return None

    def _get_choice_id_from_name(self, name: str, choices: List[Dict[str, Any]]) -> Optional[str]:
        choice_info = first_true(choices, pred=lambda e: self.facade.get_choice_name_en(e) == name)
        if choice_info is not None:
            return choice_info["choice_id"]
        else:
            return None

    @staticmethod
    def _get_data_holding_type_from_data(data: FullAnnotationData) -> AnnotationDataHoldingType:
        if data["_type"] in ["Segmentation", "SegmentationV2"]:
            return AnnotationDataHoldingType.OUTER
        else:
            return AnnotationDataHoldingType.INNER

    @staticmethod
    def _create_annotation_id(data: FullAnnotationData, label_id: str) -> str:
        if data["_type"] == "Classification":
            return label_id
        else:
            return str(uuid.uuid4())

    def _to_additional_data_list(self, attributes: Dict[str, Any], label_info: LabelV1) -> List[AdditionalData]:
        additional_data_list: List[AdditionalData] = []
        for key, value in attributes.items():
            specs_additional_data = self._get_additional_data_from_attribute_name(key, label_info)
            if specs_additional_data is None:
                logger.warning(f"attribute_name={key} が存在しません。")
                continue

            additional_data = AdditionalData(
                additional_data_definition_id=specs_additional_data["additional_data_definition_id"],
                flag=None,
                integer=None,
                choice=None,
                comment=None,
            )
            additional_data_type = AdditionalDataDefinitionType(specs_additional_data["type"])
            if additional_data_type == AdditionalDataDefinitionType.FLAG:
                additional_data.flag = value
            elif additional_data_type == AdditionalDataDefinitionType.INTEGER:
                additional_data.integer = value
            elif additional_data_type in [
                AdditionalDataDefinitionType.TEXT,
                AdditionalDataDefinitionType.COMMENT,
                AdditionalDataDefinitionType.TRACKING,
                AdditionalDataDefinitionType.LINK,
            ]:
                additional_data.comment = value
            elif additional_data_type in [AdditionalDataDefinitionType.CHOICE, AdditionalDataDefinitionType.SELECT]:
                additional_data.choice = self._get_choice_id_from_name(value, specs_additional_data["choices"])
            else:
                logger.warning(f"additional_data_type={additional_data_type}が不正です。")
                continue

            additional_data_list.append(additional_data)

        return additional_data_list

    def _to_annotation_detail_for_request(
        self, project_id: str, parser: SimpleAnnotationParser, detail: ImportedSimpleAnnotationDetail
    ) -> Optional[AnnotationDetail]:
        """
        Request Bodyに渡すDataClassに変換する。塗りつぶし画像があれば、それをS3にアップロードする。

        Args:
            project_id:
            parser:
            detail:

        Returns:

        """
        label_info = self.get_label_info_from_label_name(detail.label)
        if label_info is None:
            return None

        additional_data_list: List[AdditionalData] = self._to_additional_data_list(detail.attributes, label_info)
        data_holding_type = self._get_data_holding_type_from_data(detail.data)

        dest_obj = AnnotationDetail(
            label_id=label_info["label_id"],
            annotation_id=detail.annotation_id if detail.annotation_id is not None else str(uuid.uuid4()),
            account_id=self.service.api.account_id,
            data_holding_type=data_holding_type,
            data=detail.data,
            additional_data_list=additional_data_list,
            is_protected=False,
            etag=None,
            url=None,
            path=None,
            created_datetime=None,
            updated_datetime=None,
        )

        if data_holding_type == AnnotationDataHoldingType.OUTER:
            data_uri = detail.data["data_uri"]
            with parser.open_outer_file(data_uri) as f:
                s3_path = self.service.wrapper.upload_data_to_s3(project_id, f, content_type="image/png")
                dest_obj.path = s3_path
                logger.debug(f"{parser.task_id}/{parser.input_data_id}/{data_uri} をS3にアップロードしました。")

        return dest_obj

    def parser_to_request_body(
        self,
        project_id: str,
        parser: SimpleAnnotationParser,
        details: List[ImportedSimpleAnnotationDetail],
        old_annotation: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        request_details: List[Dict[str, Any]] = []
        for detail in details:
            request_detail = self._to_annotation_detail_for_request(project_id, parser, detail)

            if request_detail is not None:
                # Enumをシリアライズするため、一度JSONにしてからDictに変換する
                request_details.append(json.loads(request_detail.to_json()))

        updated_datetime = old_annotation["updated_datetime"] if old_annotation is not None else None

        request_body = {
            "project_id": project_id,
            "task_id": parser.task_id,
            "input_data_id": parser.input_data_id,
            "details": request_details,
            "updated_datetime": updated_datetime,
        }

        return request_body

    def put_annotation_for_input_data(
        self, project_id: str, parser: SimpleAnnotationParser, overwrite: bool = False
    ) -> bool:

        task_id = parser.task_id
        input_data_id = parser.input_data_id

        simple_annotation: ImportedSimpleAnnotation = ImportedSimpleAnnotation.from_dict(parser.load_json())
        if len(simple_annotation.details) == 0:
            logger.debug(
                f"task_id={task_id}, input_data_id={input_data_id} : インポート元にアノテーションデータがないため、アノテーションの登録をスキップします。"
            )
            return False

        input_data = self.service.wrapper.get_input_data_or_none(project_id, input_data_id)
        if input_data is None:
            logger.warning(f"task_id= '{task_id}, input_data_id = '{input_data_id}' は存在しません。")
            return False

        old_annotation, _ = self.service.api.get_editor_annotation(project_id, task_id, input_data_id)
        if len(old_annotation["details"]) > 0 and not overwrite:
            logger.debug(
                f"task_id={task_id}, input_data_id={input_data_id} : "
                f"インポート先のタスクに既にアノテーションが存在するため、アノテーションの登録をスキップします。"
            )
            return False

        logger.info(f"task_id={task_id}, input_data_id={input_data_id} : アノテーションを登録します。")
        request_body = self.parser_to_request_body(
            project_id, parser, simple_annotation.details, old_annotation=old_annotation
        )

        self.service.api.put_annotation(project_id, task_id, input_data_id, request_body=request_body)
        return True

    def put_annotation_for_task(
        self, project_id: str, task_parser: SimpleAnnotationParserByTask, overwrite: bool
    ) -> int:

        logger.info(f"タスク'{task_parser.task_id}'に対してアノテーションを登録します。")

        success_count = 0
        for parser in task_parser.lazy_parse():
            try:
                if self.put_annotation_for_input_data(project_id, parser, overwrite=overwrite):
                    success_count += 1
            except Exception as e:  # pylint: disable=broad-except
                logger.warning(
                    f"task_id={parser.task_id}, input_data_id={parser.input_data_id} のアノテーションインポートに失敗しました。: {e}"
                )

        logger.info(f"タスク'{task_parser.task_id}'の入力データ {success_count} 個に対してアノテーションをインポートしました。")
        return success_count

    def execute_task(
        self, project_id: str, task_parser: SimpleAnnotationParserByTask, my_account_id: str, overwrite: bool = False
    ) -> bool:
        """
        1個のタスクに対してアノテーションを登録する。

        Args:
            project_id:
            task_parser:
            my_account_id: 自分自身のアカウントID
            overwrite:

        Returns:
            1個以上の入力データのアノテーションを変更したか

        """
        task_id = task_parser.task_id
        if not self.confirm_processing(f"task_id={task_id} のアノテーションをインポートしますか？"):
            return False

        logger.info(f"task_id={task_id} に対して処理します。")

        task = self.service.wrapper.get_task_or_none(project_id, task_id)
        if task is None:
            logger.warning(f"task_id = '{task_id}' は存在しません。")
            return False

        if task["status"] in [TaskStatus.WORKING.value, TaskStatus.COMPLETE.value]:
            logger.info(f"タスク'{task_id}'は作業中または受入完了状態のため、インポートをスキップします。 status={task['status']}")
            return False

        if not can_put_annotation(task, my_account_id):
            logger.debug(f"タスク'{task_id}'は、過去に誰かに割り当てられたタスクで、現在の担当者が自分自身でないため、アノテーションのインポートをスキップします。")
            return False

        result_count = self.put_annotation_for_task(project_id, task_parser, overwrite)
        return result_count > 0

    @staticmethod
    def validate(args: argparse.Namespace) -> bool:
        COMMON_MESSAGE = "annofabcli annotation import: error:"
        annotation_path = Path(args.annotation)
        if not annotation_path.exists():
            print(
                f"{COMMON_MESSAGE} argument --annotation: ZIPファイルまたはディレクトリが存在しません。'{str(annotation_path)}'",
                file=sys.stderr,
            )
            return False

        elif annotation_path.is_file() and not zipfile.is_zipfile(str(annotation_path)):
            print(f"{COMMON_MESSAGE} argument --annotation: ZIPファイルまたはディレクトリを指定してください。", file=sys.stderr)
            return False

        return True

    def main(self):
        args = self.args
        if not self.validate(args):
            return

        project_id = args.project_id
        annotation_path = Path(args.annotation)

        super().validate_project(project_id, [ProjectMemberRole.OWNER])

        task_id_list = annofabcli.common.cli.get_list_from_args(args.task_id)
        my_account_id = self.service.api.account_id

        # Simpleアノテーションの読み込み
        if annotation_path.is_file():
            iter_task_parser = lazy_parse_simple_annotation_zip_by_task(annotation_path)
        else:
            iter_task_parser = lazy_parse_simple_annotation_dir_by_task(annotation_path)

        success_count = 0
        for task_parser in iter_task_parser:
            try:
                if len(task_id_list) > 0:
                    # コマンドライン引数で --task_idが指定された場合は、対象のタスクのみインポートする
                    if task_parser.task_id in task_id_list:
                        if self.execute_task(
                            project_id, task_parser, overwrite=args.overwrite, my_account_id=my_account_id
                        ):
                            success_count += 1
                else:
                    # コマンドライン引数で --task_idが指定されていない場合はすべてをインポートする
                    if self.execute_task(
                        project_id, task_parser, overwrite=args.overwrite, my_account_id=my_account_id
                    ):
                        success_count += 1

            except Exception as e:  # pylint: disable=broad-except
                logger.warning(f"task_id={task_parser.task_id} のアノテーションインポートに失敗しました。: {e}")

        logger.info(f"{success_count} 個のタスクに対してアノテーションをインポートしました。")


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    ImportAnnotation(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    argument_parser = ArgumentParser(parser)

    argument_parser.add_project_id()

    parser.add_argument(
        "--annotation",
        type=str,
        required=True,
        help="Simpleアノテーションと同じフォルダ構成のzipファイル or ディレクトリのパスを指定してください。" "タスクの状態が作業中/完了の場合はインポートしません。",
    )

    argument_parser.add_task_id(required=False)

    parser.add_argument(
        "--overwrite",
        action="store_true",
        help="指定した場合、すでに存在するアノテーションを上書きします（入力データ単位）。" "指定しなければ、アノテーションのインポートをスキップします。",
    )

    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "import"
    subcommand_help = "アノテーションをインポートします。"
    description = (
        "アノテーションをインポートします。"
        "アノテーションのフォーマットは、Simpleアノテーション(v2)と同じフォルダ構成のzipファイルまたはディレクトリです。"
        "ただし、作業中/完了状態のタスク、または「過去に割り当てられていて現在の担当者が自分自身でない」タスクはリストアできません。"
    )
    epilog = "チェッカーまたはオーナロールを持つユーザで実行してください。"

    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description, epilog=epilog)
    parse_args(parser)



from annofabcli.common import cli, enums, exceptions, typing, utils
from annofabcli.common.facade import AnnofabApiFacade

from .__version__ import __version__

__all__ = [
    "AnnofabApiFacade",
    "cli",
    "enums",
    "exceptions",
    "typing",
    "utils",
    "__version__",
]



import argparse
import logging
from typing import Any, Dict, List

import requests
from annofabapi.models import ProjectMemberRole

import annofabcli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import AbstractCommandLineInterface, ArgumentParser, build_annofabapi_resource_and_login

logger = logging.getLogger(__name__)


class DeleteInputData(AbstractCommandLineInterface):
    """
    タスクに使われていない入力データを削除する。
    """

    @annofabcli.utils.allow_404_error
    def get_input_data(self, project_id: str, input_data_id: str) -> Dict[str, Any]:
        input_data, _ = self.service.api.get_input_data(project_id, input_data_id)
        return input_data

    def confirm_delete_input_data(self, input_data_id: str, input_data_name: str) -> bool:
        message_for_confirm = (
            f"入力データ(input_data_id='{input_data_id}', " f"input_data_name='{input_data_name}') を削除しますか？"
        )
        return self.confirm_processing(message_for_confirm)

    def delete_input_data(self, project_id: str, input_data_id: str, input_data_index: int):
        input_data = self.get_input_data(project_id, input_data_id)
        if input_data is None:
            logger.info(f"input_data_id={input_data_id} は存在しません。")
            return False

        task_list = self.service.wrapper.get_all_tasks(project_id, query_params={"input_data_ids": input_data_id})
        input_data_name = input_data["input_data_name"]
        if len(task_list) > 0:
            task_id_list = [e["task_id"] for e in task_list]
            logger.info(
                f"入力データ(input_data_id='{input_data_id}', "
                f"input_data_name='{input_data_name}')はタスクに使われているので、削除しません。"
                f"task_id_list='{task_id_list}'"
            )
            return False

        if not self.confirm_delete_input_data(input_data_id, input_data_name):
            return False

        self.service.api.delete_input_data(project_id, input_data_id)
        logger.info(
            f"{str(input_data_index+1)} 件目: 入力データ(input_data_id='{input_data_id}', "
            f"input_data_name='{input_data_name}') を削除しました。"
        )
        return True

    def delete_input_data_list(self, project_id: str, input_data_id_list: List[str]):
        """
        タスクに使われていない入力データを削除する。
        """

        super().validate_project(project_id, [ProjectMemberRole.OWNER])
        project_title = self.facade.get_project_title(project_id)
        logger.info(f"プロジェクト'{project_title}'から 、{len(input_data_id_list)} 件の入力データを削除します。")

        count_delete_input_data = 0
        for input_data_index, input_data_id in enumerate(input_data_id_list):
            try:
                result = self.delete_input_data(project_id, input_data_id, input_data_index=input_data_index)
                if result:
                    count_delete_input_data += 1

            except requests.exceptions.HTTPError as e:
                logger.warning(e)
                logger.warning(f"input_data_id='{input_data_id}'の削除に失敗しました。")
                continue

        logger.info(f"プロジェクト'{project_title}'から 、{count_delete_input_data} 件の入力データを削除しました。")

    def main(self):
        args = self.args
        input_data_id_list = annofabcli.common.cli.get_list_from_args(args.input_data_id)
        self.delete_input_data_list(args.project_id, input_data_id_list=input_data_id_list)


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    DeleteInputData(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    argument_parser = ArgumentParser(parser)

    argument_parser.add_project_id()
    parser.add_argument(
        "-i",
        "--input_data_id",
        type=str,
        required=True,
        nargs="+",
        help="削除対象の入力データのinput_data_idを指定します。" "`file://`を先頭に付けると、input_data_idの一覧が記載されたファイルを指定できます。",
    )

    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "delete"
    subcommand_help = "入力データを削除します。"
    description = "入力データを削除します。ただし、タスクに使われている入力データは削除できません。"
    epilog = "オーナロールを持つユーザで実行してください。"

    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description, epilog=epilog)
    parse_args(parser)



import argparse
import logging

import pandas

import annofabcli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import AbstractCommandLineInterface, ArgumentParser, build_annofabapi_resource_and_login
from annofabcli.statistics.csv import Csv
from annofabcli.statistics.table import Table

logger = logging.getLogger(__name__)


class WritePerfomancePerUser(AbstractCommandLineInterface):
    def main(self):
        args = self.args
        csv_obj = Csv(outdir=args.output_dir)

        df_task = pandas.read_csv(args.task_csv)
        df_task_history = pandas.read_csv(args.task_history_csv)
        df_labor = pandas.read_csv(args.labor_csv)

        df_count_ratio = Table.create_annotation_count_ratio_df(df_task_history, df_task)

        df_perfomance_per_user = Table.create_productivity_per_user_from_aw_time(
            df_task_history, df_labor, df_count_ratio
        )

        csv_obj.write_productivity_per_user(df_perfomance_per_user)


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    WritePerfomancePerUser(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    argument_parser = ArgumentParser(parser)

    argument_parser.add_project_id()
    parser.add_argument(
        "--task_csv",
        type=str,
        required=True,
        help=("タスク一覧のCSVファイルのパスを指定してください。" "CSVは、'statistics visualize'コマンドの出力結果である'タスクlist.csv'と同じフォーマットです。"),
    )
    parser.add_argument(
        "--task_history_csv",
        type=str,
        required=True,
        help=("タスク履歴一覧のCSVファイルのパスを指定してください。" "CSVは、'statistics visualize'コマンドの出力結果である'タスク履歴list.csv'と同じフォーマットです。"),
    )
    parser.add_argument(
        "--labor_csv",
        type=str,
        required=True,
        help=("労務管理一覧のCSVファイルのパスを指定してください。" "CSVは、'statistics visualize'コマンドの出力結果である'労務管理list.csv'と同じフォーマットです。"),
    )
    parser.add_argument("-o", "--output_dir", type=str, required=True, help="出力ディレクトリのパス")

    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "write_peformance_csv_per_user"
    subcommand_help = "ユーザごとのパフォーマンス（生産性や品質）をCSVに出力します。"
    description = "ユーザごとのパフォーマンス（生産性や品質）をCSVに出力します。"
    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description)
    parse_args(parser)



import argparse

import annofabcli
import annofabcli.common.cli
import annofabcli.task.cancel_acceptance
import annofabcli.task.change_operator
import annofabcli.task.complete_tasks
import annofabcli.task.delete_tasks
import annofabcli.task.list_tasks
import annofabcli.task.list_tasks_added_task_history
import annofabcli.task.list_tasks_with_json
import annofabcli.task.put_tasks
import annofabcli.task.reject_tasks
import annofabcli.task.update_metadata_of_task


def parse_args(parser: argparse.ArgumentParser):

    subparsers = parser.add_subparsers(dest="subcommand_name")

    # サブコマンドの定義
    annofabcli.task.cancel_acceptance.add_parser(subparsers)
    annofabcli.task.change_operator.add_parser(subparsers)
    annofabcli.task.complete_tasks.add_parser(subparsers)
    annofabcli.task.delete_tasks.add_parser(subparsers)
    annofabcli.task.list_tasks.add_parser(subparsers)
    annofabcli.task.list_tasks_with_json.add_parser(subparsers)
    annofabcli.task.list_tasks_added_task_history.add_parser(subparsers)
    annofabcli.task.put_tasks.add_parser(subparsers)
    annofabcli.task.reject_tasks.add_parser(subparsers)
    annofabcli.task.update_metadata_of_task.add_parser(subparsers)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "task"
    subcommand_help = "タスク関係のサブコマンド"
    description = "タスク関係のサブコマンド"

    parser = annofabcli.common.cli.add_parser(
        subparsers, subcommand_name, subcommand_help, description, is_subcommand=False
    )
    parse_args(parser)



import argparse
import copy
import logging
from typing import Any, Dict, List, Optional

from annofabapi.models import JobInfo, JobType

import annofabcli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import AbstractCommandLineInterface, ArgumentParser, build_annofabapi_resource_and_login
from annofabcli.common.enums import FormatArgument

logger = logging.getLogger(__name__)


class ListJob(AbstractCommandLineInterface):
    """
    ジョブ一覧を表示する。
    """

    def get_job_list(
        self, project_id: str, job_type: JobType, job_query: Optional[Dict[str, Any]] = None
    ) -> List[JobInfo]:
        """
        ジョブ一覧を取得する。
        """

        if job_query is not None:
            query_params = copy.deepcopy(job_query)
        else:
            query_params = {}

        query_params["type"] = job_type.value

        logger.debug(f"query_params: {query_params}")
        job_list = self.service.wrapper.get_all_project_job(project_id, query_params=query_params)
        return job_list

    def print_job_list(self, project_id: str, job_type: JobType, job_query: Optional[Dict[str, Any]] = None):
        """
        ジョブ一覧を出力する

        Args:
            project_id: 対象のproject_id
            job_type: ジョブタイプ
            job_query:

        """

        super().validate_project(project_id, project_member_roles=None)

        job_list = self.get_job_list(project_id, job_type=job_type, job_query=job_query)
        logger.info(f"ジョブ一覧の件数: {len(job_list)}")
        self.print_according_to_format(job_list)

    def main(self):
        args = self.args
        # job_query = annofabcli.common.cli.get_json_from_args(args.job_query)
        job_type = JobType(args.job_type)
        self.print_job_list(args.project_id, job_type=job_type, job_query=None)


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    ListJob(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    argument_parser = ArgumentParser(parser)

    job_choices = [e.value for e in JobType]
    argument_parser.add_project_id()

    parser.add_argument("--job_type", type=str, choices=job_choices, required=True, help="ジョブタイプを指定します。")

    # クエリがうまく動かないので、コメントアウトする
    # parser.add_argument(
    #     '--job_query', type=str, help='ジョブの検索クエリをJSON形式で指定します。指定しない場合は、最新のジョブを1個取得します。 '
    #     '`file://`を先頭に付けると、JSON形式のファイルを指定できます。'
    #     '`limit` キーを指定できます。')

    argument_parser.add_format(
        choices=[FormatArgument.CSV, FormatArgument.JSON, FormatArgument.PRETTY_JSON], default=FormatArgument.CSV
    )
    argument_parser.add_output()
    argument_parser.add_csv_format()

    argument_parser.add_query()
    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "list"
    subcommand_help = "ジョブ一覧を出力します。"
    description = "ジョブ一覧を出力します。"

    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description)
    parse_args(parser)



"""
検査コメント一覧を出力する。
"""

import argparse
import json
import logging
from pathlib import Path
from typing import Callable, List, Optional

import annofabapi
from annofabapi.models import Inspection

import annofabcli
import annofabcli.common.cli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import AbstractCommandLineInterface, ArgumentParser, build_annofabapi_resource_and_login
from annofabcli.common.download import DownloadingFile
from annofabcli.common.enums import FormatArgument
from annofabcli.common.visualize import AddProps
from annofabcli.inspection_comment.list_inspections import create_filter_func

logger = logging.getLogger(__name__)

FilterInspectionFunc = Callable[[Inspection], bool]


class ListInspectionCommentWithJsonMain:
    def __init__(self, service: annofabapi.Resource):
        self.service = service
        self.facade = AnnofabApiFacade(service)

    def filter_inspection_list(
        self,
        project_id: str,
        inspection_comment_list: List[Inspection],
        task_id_list: Optional[List[str]] = None,
        filter_inspection_comment: Optional[FilterInspectionFunc] = None,
    ) -> List[Inspection]:
        """
        引数の検査コメント一覧に`commenter_username`など、ユーザが知りたい情報を追加する。

        Args:
            inspection_list: 検査コメント一覧
            filter_inspection: 検索コメントを絞り込むための関数

        Returns:
            情報が追加された検査コメント一覧
        """

        def filter_task_id(e):
            if task_id_list is None or len(task_id_list) == 0:
                return True
            return e["task_id"] in task_id_list

        def filter_local_inspection_comment(e):
            if filter_inspection_comment is None:
                return True
            return filter_inspection_comment(e)

        inspection_list = [
            e for e in inspection_comment_list if filter_local_inspection_comment(e) and filter_task_id(e)
        ]
        visualize = AddProps(self.service, project_id)
        return [visualize.add_properties_to_inspection(e) for e in inspection_list]

    def get_inspection_comment_list(
        self,
        project_id: str,
        inspection_comment_json: Optional[Path],
        task_id_list: Optional[List[str]],
        only_reply: bool,
        exclude_reply: bool,
    ) -> List[Inspection]:

        if inspection_comment_json is None:
            downloading_obj = DownloadingFile(self.service)
            cache_dir = annofabcli.utils.get_cache_dir()
            json_path = cache_dir / f"{project_id}-inspection.json"

            downloading_obj.download_inspection_json(project_id, str(json_path))
        else:
            json_path = inspection_comment_json

        filter_inspection_comment = create_filter_func(only_reply=only_reply, exclude_reply=exclude_reply)
        with json_path.open() as f:
            inspection_comment_list = json.load(f)

        return self.filter_inspection_list(
            project_id,
            inspection_comment_list=inspection_comment_list,
            task_id_list=task_id_list,
            filter_inspection_comment=filter_inspection_comment,
        )


class ListInspectionCommentWithJson(AbstractCommandLineInterface):
    def main(self):
        args = self.args
        project_id = args.project_id
        super().validate_project(project_id, project_member_roles=None)

        task_id_list = annofabcli.common.cli.get_list_from_args(args.task_id) if args.task_id is not None else None
        main_obj = ListInspectionCommentWithJsonMain(self.service)
        inspection_comment_list = main_obj.get_inspection_comment_list(
            project_id=project_id,
            inspection_comment_json=args.inspection_comment_json,
            task_id_list=task_id_list,
            exclude_reply=args.exclude_reply,
            only_reply=args.only_reply,
        )

        logger.info(f"検査コメントの件数: {len(inspection_comment_list)}")

        self.print_according_to_format(inspection_comment_list)


def parse_args(parser: argparse.ArgumentParser):
    argument_parser = ArgumentParser(parser)

    argument_parser.add_project_id()
    argument_parser.add_task_id(
        required=False,
        help_message="対象のタスクのtask_idを指定します。　"
        "`--inspection_comment_json`を指定しないときは、必須です。"
        "`file://`を先頭に付けると、task_idの一覧が記載されたファイルを指定できます。",
    )

    reply_comment_group = parser.add_mutually_exclusive_group()
    reply_comment_group.add_argument("--only_reply", action="store_true", help="返信コメントのみを出力する。")
    reply_comment_group.add_argument("--exclude_reply", action="store_true", help="返信コメントを除外して出力する。")

    parser.add_argument(
        "--inspection_comment_json",
        type=str,
        help="検査コメント情報が記載されたJSONファイルのパスを指定すると、JSONに記載された情報を元に検査コメント一覧を出力します。指定しない場合、全件ファイルをダウンロードします。"
        "JSONファイルは`$ annofabcli project download inspection_comment`コマンドで取得できます。",
    )

    argument_parser.add_format(
        choices=[
            FormatArgument.CSV,
            FormatArgument.JSON,
            FormatArgument.PRETTY_JSON,
            FormatArgument.INSPECTION_ID_LIST,
        ],
        default=FormatArgument.CSV,
    )
    argument_parser.add_output()
    argument_parser.add_csv_format()

    parser.set_defaults(subcommand_func=main)


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    ListInspectionCommentWithJson(service, facade, args).main()


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "list_with_json"

    subcommand_help = "検査コメント全件ファイルから一覧を出力します。"

    description = "検査コメント全件ファイルから一覧を出力します。"

    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description)
    parse_args(parser)



import argparse
import logging
import sys
import uuid
import zipfile
from dataclasses import dataclass
from distutils.util import strtobool  # pylint: disable=import-error,no-name-in-module
from functools import partial
from multiprocessing import Pool
from pathlib import Path
from typing import Any, Dict, List, Optional

import annofabapi
import pandas
import requests
from annofabapi.models import JobType, ProjectMemberRole
from dataclasses_json import DataClassJsonMixin

import annofabcli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import (
    AbstractCommandLineInterface,
    ArgumentParser,
    build_annofabapi_resource_and_login,
    get_json_from_args,
    get_wait_options_from_args,
    prompt_yesnoall,
)
from annofabcli.common.dataclasses import WaitOptions
from annofabcli.common.utils import get_file_scheme_path

logger = logging.getLogger(__name__)

DEFAULT_WAIT_OPTIONS = WaitOptions(interval=60, max_tries=360)


@dataclass
class CsvInputData(DataClassJsonMixin):
    """
    CSVに記載されている入力データ
    """

    input_data_name: str
    input_data_path: str
    input_data_id: str
    sign_required: Optional[bool]


class SubPutInputData:
    """
    1個の入力データを登録するためのクラス。multiprocessing.Pool対応。

    Args:
        service:
        facade:
        all_yes:
    """

    def __init__(self, service: annofabapi.Resource, facade: AnnofabApiFacade, all_yes: bool = False):
        self.service = service
        self.facade = facade
        self.all_yes = all_yes

    def put_input_data(
        self, project_id: str, csv_input_data: CsvInputData, last_updated_datetime: Optional[str] = None
    ):

        request_body: Dict[str, Any] = {"last_updated_datetime": last_updated_datetime}

        file_path = get_file_scheme_path(csv_input_data.input_data_path)
        if file_path is not None:
            request_body.update(
                {"input_data_name": csv_input_data.input_data_name, "sign_required": csv_input_data.sign_required}
            )
            logger.debug(f"'{file_path}'を入力データとして登録します。input_data_name={csv_input_data.input_data_name}")
            self.service.wrapper.put_input_data_from_file(
                project_id, input_data_id=csv_input_data.input_data_id, file_path=file_path, request_body=request_body
            )

        else:
            request_body.update(
                {
                    "input_data_name": csv_input_data.input_data_name,
                    "input_data_path": csv_input_data.input_data_path,
                    "sign_required": csv_input_data.sign_required,
                }
            )

            self.service.api.put_input_data(project_id, csv_input_data.input_data_id, request_body=request_body)

    def confirm_processing(self, confirm_message: str) -> bool:
        """
        `all_yes`属性を見て、処理するかどうかユーザに問い合わせる。
        "ALL"が入力されたら、`all_yes`属性をTrueにする

        Args:
            task_id: 処理するtask_id
            confirm_message: 確認メッセージ

        Returns:
            True: Yes, False: No

        """
        if self.all_yes:
            return True

        yes, all_yes = prompt_yesnoall(confirm_message)

        if all_yes:
            self.all_yes = True

        return yes

    def confirm_put_input_data(self, csv_input_data: CsvInputData, already_exists: bool = False) -> bool:
        message_for_confirm = f"input_data_name='{csv_input_data.input_data_name}' の入力データを登録しますか？"
        if already_exists:
            message_for_confirm += f"input_data_id={csv_input_data.input_data_id} を上書きします。"
        return self.confirm_processing(message_for_confirm)

    def put_input_data_main(self, project_id: str, csv_input_data: CsvInputData, overwrite: bool = False) -> bool:
        last_updated_datetime = None
        input_data_id = csv_input_data.input_data_id
        input_data_path = csv_input_data.input_data_path
        input_data = self.service.wrapper.get_input_data_or_none(project_id, input_data_id)

        if input_data is not None:
            if overwrite:
                logger.debug(f"input_data_id={input_data_id} はすでに存在します。")
                last_updated_datetime = input_data["updated_datetime"]
            else:
                logger.debug(f"input_data_id={input_data_id} がすでに存在するのでスキップします。")
                return False

        file_path = get_file_scheme_path(input_data_path)
        logger.debug(f"csv_input_data={csv_input_data}")
        if file_path is not None:
            if not Path(file_path).exists():
                logger.warning(f"{input_data_path} は存在しません。")
                return False

        if not self.confirm_put_input_data(csv_input_data, already_exists=(last_updated_datetime is not None)):
            return False

        # 入力データを登録
        try:
            self.put_input_data(project_id, csv_input_data, last_updated_datetime=last_updated_datetime)
            logger.debug(
                f"入力データを登録しました。"
                f"input_data_id={csv_input_data.input_data_id}, "
                f"input_data_name={csv_input_data.input_data_name}"
            )
            return True

        except requests.exceptions.HTTPError as e:
            logger.warning(e)
            logger.warning(
                f"入力データの登録に失敗しました。"
                f"input_data_id={csv_input_data.input_data_id}, "
                f"input_data_name={csv_input_data.input_data_name}"
            )
            return False


class PutInputData(AbstractCommandLineInterface):
    """
    入力データをCSVで登録する。
    """

    def put_input_data_list(
        self,
        project_id: str,
        input_data_list: List[CsvInputData],
        overwrite: bool = False,
        parallelism: Optional[int] = None,
    ) -> None:
        """
        入力データを一括で登録する。

        Args:
            project_id: 入力データの登録先プロジェクトのプロジェクトID
            input_data_list: 入力データList
            overwrite: Trueならば、input_data_idがすでに存在していたら上書きします。Falseならばスキップします。
            parallelism: 並列度

        """

        project_title = self.facade.get_project_title(project_id)
        logger.info(f"{project_title} に、{len(input_data_list)} 件の入力データを登録します。")

        count_put_input_data = 0

        obj = SubPutInputData(service=self.service, facade=self.facade, all_yes=self.all_yes)
        if parallelism is not None:
            partial_func = partial(obj.put_input_data_main, project_id, overwrite=overwrite)
            with Pool(parallelism) as pool:
                result_bool_list = pool.map(partial_func, input_data_list)
                count_put_input_data = len([e for e in result_bool_list if e])

        else:
            for csv_input_data in input_data_list:
                result = obj.put_input_data_main(project_id, csv_input_data=csv_input_data, overwrite=overwrite)
                if result:
                    count_put_input_data += 1

        logger.info(f"{project_title} に、{count_put_input_data} / {len(input_data_list)} 件の入力データを登録しました。")

    @staticmethod
    def get_input_data_list_from_csv(csv_path: Path) -> List[CsvInputData]:
        def create_input_data(e):
            input_data_id = e.input_data_id if not pandas.isna(e.input_data_id) else str(uuid.uuid4())
            sign_required = bool(strtobool(str(e.sign_required))) if not pandas.isna(e.sign_required) else None
            return CsvInputData(
                input_data_name=e.input_data_name,
                input_data_path=e.input_data_path,
                input_data_id=input_data_id,
                sign_required=sign_required,
            )

        df = pandas.read_csv(
            str(csv_path),
            sep=",",
            header=None,
            names=("input_data_name", "input_data_path", "input_data_id", "sign_required"),
        )
        df_duplicated_input_data_name = df[df["input_data_name"].duplicated()]
        if len(df_duplicated_input_data_name) > 0:
            logger.error(
                f"{csv_path}に記載されている`input_data_name`が重複しています。\n"
                f"{df_duplicated_input_data_name['input_data_name'].to_string(index=False)}"
            )
            raise RuntimeError(f"{csv_path}に記載されている`input_data_name`が重複しています。")

        df_duplicated_input_data_path = df[df["input_data_path"].duplicated()]
        if len(df_duplicated_input_data_path) > 0:
            logger.error(
                f"{csv_path}に記載されている`input_data_path`が重複しています。\n"
                f"{df_duplicated_input_data_name['input_data_path'].to_string(index=False)}"
            )
            raise RuntimeError(f"{csv_path}に記載されている`input_data_path`が重複しています。")

        input_data_list = [create_input_data(e) for e in df.itertuples()]

        return input_data_list

    def put_input_data_from_zip_file(
        self,
        project_id: str,
        zip_file: Path,
        wait_options: WaitOptions,
        input_data_name_for_zip: Optional[str] = None,
        wait: bool = False,
    ) -> None:
        """
        zipファイルを入力データとして登録する

        Args:
            project_id: 入力データの登録先プロジェクトのプロジェクトID
            zip_file: 入力データとして登録するzipファイルのパス
            input_data_name_for_zip: zipファイルのinput_data_name
            wait: 入力データの登録が完了するまで待つかどうか

        """

        project_title = self.facade.get_project_title(project_id)
        logger.info(f"{project_title} に、{str(zip_file)} を登録します。")

        request_body = {}
        if input_data_name_for_zip is not None:
            request_body["input_data_name"] = input_data_name_for_zip

        self.service.wrapper.put_input_data_from_file(
            project_id,
            input_data_id=str(uuid.uuid4()),
            file_path=str(zip_file),
            content_type="application/zip",
            request_body=request_body,
        )
        logger.info(f"入力データの登録中です（サーバ側の処理）。")

        if wait:
            MAX_WAIT_MINUTUE = wait_options.max_tries * wait_options.interval / 60
            logger.info(f"最大{MAX_WAIT_MINUTUE}分間、処理が終了するまで待ちます。")

            result = self.service.wrapper.wait_for_completion(
                project_id,
                job_type=JobType.GEN_INPUTS,
                job_access_interval=wait_options.interval,
                max_job_access=wait_options.max_tries,
            )
            if result:
                logger.info(f"入力データの登録が完了しました。")
            else:
                logger.warning(f"入力データの登録に失敗しました。または、{MAX_WAIT_MINUTUE}分間待っても、入力データの登録が完了しませんでした。")

    @staticmethod
    def validate(args: argparse.Namespace) -> bool:
        COMMON_MESSAGE = "annofabcli input_data put: error:"
        if args.zip is not None:
            if not Path(args.zip).exists():
                print(f"{COMMON_MESSAGE} argument --zip: ファイルパスが存在しません。 '{args.zip}'", file=sys.stderr)
                return False

            if not zipfile.is_zipfile(args.zip):
                print(f"{COMMON_MESSAGE} argument --zip: zipファイルではありません。 '{args.zip}'", file=sys.stderr)
                return False

            if args.overwrite:
                logger.warning(f"'--zip'オプションを指定しているとき、'--overwrite'オプションは無視されます。")

            if args.parallelism is not None:
                logger.warning(f"'--zip'オプションを指定しているとき、'--parallelism'オプションは無視されます。")

        if args.csv is not None:
            if not Path(args.csv).exists():
                print(f"{COMMON_MESSAGE} argument --csv: ファイルパスが存在しません。 '{args.csv}'", file=sys.stderr)
                return False

            if args.wait:
                logger.warning(f"'--csv'オプションを指定しているとき、'--wait'オプションは無視されます。")

            if args.input_data_name_for_zip:
                logger.warning(f"'--csv'オプションを指定しているとき、'--input_data_name_for_zip'オプションは無視されます。")

            if args.parallelism is not None and not args.yes:
                print(
                    f"{COMMON_MESSAGE} argument --parallelism: '--parallelism'を指定するときは、必ず'--yes'を指定してください。",
                    file=sys.stderr,
                )
                return False

        return True

    def main(self):
        args = self.args
        if not self.validate(args):
            return

        project_id = args.project_id
        super().validate_project(project_id, [ProjectMemberRole.OWNER])

        if args.csv is not None:
            input_data_list = self.get_input_data_list_from_csv(Path(args.csv))
            self.put_input_data_list(
                project_id, input_data_list=input_data_list, overwrite=args.overwrite, parallelism=args.parallelism
            )

        elif args.zip is not None:
            wait_options = get_wait_options_from_args(get_json_from_args(args.wait_options), DEFAULT_WAIT_OPTIONS)
            self.put_input_data_from_zip_file(
                project_id,
                zip_file=Path(args.zip),
                input_data_name_for_zip=args.input_data_name_for_zip,
                wait=args.wait,
                wait_options=wait_options,
            )

        else:
            print(f"引数が不正です。", file=sys.stderr)


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    PutInputData(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    argument_parser = ArgumentParser(parser)

    argument_parser.add_project_id()

    file_group = parser.add_mutually_exclusive_group(required=True)
    file_group.add_argument(
        "--csv",
        type=str,
        help=(
            "入力データが記載されたCVファイルのパスを指定してください。"
            "CSVのフォーマットは、「1列目:input_data_name(required), 2列目:input_data_path(required), 3列目:input_data_id, "
            "4列目:sign_required(bool), ヘッダ行なし, カンマ区切り」です。"
            "input_data_pathの先頭が`file://`の場合、ローカルのファイルを入力データとして登録します。 "
            "input_data_idが空の場合はUUIDv4になります。"
            "各項目の詳細は `putInputData` API を参照してください。"
        ),
    )

    file_group.add_argument("--zip", type=str, help=("入力データとして登録するzipファイルのパスを指定してください。"))

    parser.add_argument(
        "--overwrite",
        action="store_true",
        help="指定した場合、input_data_idがすでに存在していたら上書きします。指定しなければ、スキップします。" "`--csv`を指定したときのみ有効なオプションです。",
    )

    parser.add_argument(
        "--input_data_name_for_zip",
        type=str,
        help="入力データとして登録するzipファイルのinput_data_nameを指定してください。省略した場合、`--zip`のパスになります。" "`--zip`を指定したときのみ有効なオプションです。",
    )

    parser.add_argument("--wait", action="store_true", help=("入力データの登録が完了するまで待ちます。" "`--zip`を指定したときのみ有効なオプションです。"))

    parser.add_argument(
        "--wait_options",
        type=str,
        help="入力データの登録が完了するまで待つ際のオプションをJSON形式で指定してください。`--wait`を指定したときのみ有効なオプションです。"
        "`file://`を先頭に付けるとjsonファイルを指定できます。"
        'デフォルとは`{"interval":60, "max_tries":360}` です。'
        "`interval`:完了したかを問い合わせる間隔[秒], "
        "`max_tires`:完了したかの問い合わせを最大何回行うか。",
    )

    parser.add_argument(
        "--parallelism", type=int, help="並列度。指定しない場合は、逐次的に処理します。" "'--csv'を指定したときのみ有効なオプションです。また、必ず'--yes'を指定してください。"
    )

    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "put"
    subcommand_help = "入力データを登録します。"
    description = "CSVに記載された入力データ情報やzipファイルを、入力データとして登録します。"
    epilog = "オーナロールを持つユーザで実行してください。"

    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description, epilog=epilog)
    parse_args(parser)



import argparse
import logging
from typing import Any, Dict, List, Optional

import annofabapi
import pandas
import requests
from annofabapi.models import OrganizationMember, Project, ProjectMemberRole, ProjectStatus
from more_itertools import first_true

import annofabcli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import AbstractCommandLineInterface, build_annofabapi_resource_and_login

logger = logging.getLogger(__name__)


def create_minimal_dataframe(project_list: List[Project]):
    """必要最小限の列であるDataFramewoを作成する"""
    df = pandas.DataFrame(project_list)
    df["last_tasks_updated_datetime"] = [e["summary"]["last_tasks_updated_datetime"] for e in project_list]
    return df[
        [
            "project_id",
            "title",
            "organization_name",
            "project_status",
            "input_data_type",
            "last_tasks_updated_datetime",
            "created_datetime",
        ]
    ]


class ChanegProjectStatusMain:
    def __init__(self, service: annofabapi.Resource):
        self.service = service
        self.facade = AnnofabApiFacade(service)

    @staticmethod
    def get_account_id_from_user_id(organization_member_list: List[OrganizationMember], user_id: str) -> Optional[str]:
        member = first_true(organization_member_list, pred=lambda e: e["user_id"] == user_id)
        if member is not None:
            return member["account_id"]
        else:
            return None

    def get_project_list_from_project_id(self, project_id_list: List[str]) -> List[Project]:
        """
        project_idからプロジェクト一覧を取得する。
        """
        project_list = []
        for project_id in project_id_list:
            project = self.service.wrapper.get_project_or_none(project_id)
            if project is None:
                logger.warning(f"project_id='{project_id}'のプロジェクトにアクセスできませんでした。")
                continue
            organization, _ = self.service.api.get_organization_of_project(project_id)
            project["organization_name"] = organization["organization_name"]
            project_list.append(project)

        return project_list

    def _modify_project_query(self, organization_name: str, project_query: Dict[str, Any]) -> Dict[str, Any]:
        """
        プロジェクト索クエリを修正する。
        ``user_id`` から ``account_id`` に変換する。
        ``except_user_id`` から ``except_account_id`` に変換する。
        ``page`` , ``limit``を削除」する

        Args:

            project_query: タスク検索クエリ（IN/OUT）

        Returns:
            修正したタスク検索クエリ

        """

        def remove_key(arg_key: str):
            if arg_key in project_query:
                logger.info(f"project_query から、`{arg_key}`　キーを削除しました。")
                project_query.pop(arg_key)

        remove_key("page")
        remove_key("limit")

        organization_member_list = self.service.wrapper.get_all_organization_members(organization_name)

        if "user_id" in project_query:
            user_id = project_query["user_id"]
            account_id = self.get_account_id_from_user_id(organization_member_list, user_id)
            if account_id is not None:
                project_query["account_id"] = account_id
            else:
                logger.warning(f"project_query に含まれている user_id: {user_id} のユーザが見つかりませんでした。")

        if "except_user_id" in project_query:
            except_user_id = project_query["except_user_id"]
            except_account_id = self.get_account_id_from_user_id(organization_member_list, except_user_id)
            if except_account_id is not None:
                project_query["except_account_id"] = except_account_id
            else:
                logger.warning(f"project_query に含まれている except_user_id: {except_user_id} のユーザが見つかりませんでした。")

        return project_query

    def change_status_for_project(self, project_id: str, status: ProjectStatus, force_suspend: bool = False) -> bool:
        project = self.service.wrapper.get_project_or_none(project_id)
        if project is None:
            logger.warning(f"project_id={project_id} のプロジェクトは存在しないので、スキップします。")
            return False

        if not self.facade.contains_any_project_member_role(project_id, [ProjectMemberRole.OWNER]):
            logger.warning(f"project_id={project_id}: オーナロールでないため、アノテーションzipを更新できません。project_title={project['title']}")
            return False

        logger.debug(f"{project['title']} のステータスを{status.value} に変更します。project_id={project_id}")
        project["status"] = status.value
        project["last_updated_datetime"] = project["updated_datetime"]
        project["force_suspend"] = force_suspend
        self.service.api.put_project(project_id, request_body=project)
        return True

    def change_status_for_project_list(
        self, project_id_list: List[str], status: ProjectStatus, force_suspend: bool = False
    ):
        """
        複数のプロジェクトに対して、プロジェクトのステータスを変更する。

        Args:
            project_id_list:
            status: 変更後のステータス
            force_suspend: Trueなら作業中タスクがある状態でも停止中にする。

        Returns:

        """
        logger.info(f"{len(project_id_list)} 件のプロジェクトのステータスを {status.value} に変更します。")
        success_count = 0
        for project_id in project_id_list:
            try:
                result = self.change_status_for_project(project_id, status=status, force_suspend=force_suspend)
                if result:
                    success_count += 1

            except requests.HTTPError as e:
                if e.response.status_code == requests.codes.conflict:
                    # 現在のプロジェクトの状態では、ステータスを変更できない場合に発生するエラー
                    logger.warning(e)
                else:
                    raise e

        logger.info(f"{success_count} 件のプロジェクトのステータスを {status.value} に変更しました。")


class ChangeProjectStatus(AbstractCommandLineInterface):
    def main(self):
        args = self.args
        project_id_list = annofabcli.common.cli.get_list_from_args(args.project_id)
        main_obj = ChanegProjectStatusMain(self.service)
        main_obj.change_status_for_project_list(
            project_id_list=project_id_list, status=ProjectStatus(args.status), force_suspend=args.force
        )


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    ChangeProjectStatus(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):

    parser.add_argument(
        "-p",
        "--project_id",
        type=str,
        required=True,
        nargs="+",
        help="対象プロジェクトのproject_idを指定します。`file://`を先頭に付けると、project_idの一覧が記載されたファイルを指定できます。",
    )

    parser.add_argument(
        "--status",
        type=str,
        choices=[ProjectStatus.ACTIVE.value, ProjectStatus.SUSPENDED.value],
        required=True,
        help="変更後のステータスを指定してください。",
    )

    parser.add_argument(
        "--force",
        action="store_true",
        help=f"`--status {ProjectStatus.SUSPENDED.value}`を指定している状態で、`--force`を指定した場合、作業中タスクが残っていても停止状態に変更します。",
    )

    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "change_status"
    subcommand_help = "プロジェクトのステータスを変更します。"
    description = "プロジェクトのステータスを変更します。"
    epilog = "オーナロールを持つユーザで実行してください。"

    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description, epilog=epilog)
    parse_args(parser)



import argparse
import logging
from typing import Any, Dict, List, Optional

import annofabapi
import pandas
from annofabapi.models import OrganizationMember, Project
from more_itertools import first_true

import annofabcli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import AbstractCommandLineInterface, ArgumentParser, build_annofabapi_resource_and_login
from annofabcli.common.enums import FormatArgument

logger = logging.getLogger(__name__)


def create_minimal_dataframe(project_list: List[Project]):
    """必要最小限の列であるDataFramewoを作成する"""
    df = pandas.DataFrame(project_list)
    df["last_tasks_updated_datetime"] = [e["summary"]["last_tasks_updated_datetime"] for e in project_list]
    return df[
        [
            "project_id",
            "title",
            "organization_name",
            "project_status",
            "input_data_type",
            "last_tasks_updated_datetime",
            "created_datetime",
        ]
    ]


class ListProjectMain:
    def __init__(self, service: annofabapi.Resource):
        self.service = service
        self.facade = AnnofabApiFacade(service)

    @staticmethod
    def get_account_id_from_user_id(organization_member_list: List[OrganizationMember], user_id: str) -> Optional[str]:
        member = first_true(organization_member_list, pred=lambda e: e["user_id"] == user_id)
        if member is not None:
            return member["account_id"]
        else:
            return None

    def get_project_list_from_project_id(self, project_id_list: List[str]) -> List[Project]:
        """
        project_idからプロジェクト一覧を取得する。
        """
        project_list = []
        for project_id in project_id_list:
            project = self.service.wrapper.get_project_or_none(project_id)
            if project is None:
                logger.warning(f"project_id='{project_id}'のプロジェクトにアクセスできませんでした。")
                continue
            organization, _ = self.service.api.get_organization_of_project(project_id)
            project["organization_name"] = organization["organization_name"]
            project_list.append(project)

        return project_list

    def _modify_project_query(self, organization_name: str, project_query: Dict[str, Any]) -> Dict[str, Any]:
        """
        プロジェクト索クエリを修正する。
        ``user_id`` から ``account_id`` に変換する。
        ``except_user_id`` から ``except_account_id`` に変換する。
        ``page`` , ``limit``を削除」する

        Args:

            project_query: タスク検索クエリ（IN/OUT）

        Returns:
            修正したタスク検索クエリ

        """

        def remove_key(arg_key: str):
            if arg_key in project_query:
                logger.info(f"project_query から、`{arg_key}`　キーを削除しました。")
                project_query.pop(arg_key)

        remove_key("page")
        remove_key("limit")

        organization_member_list = self.service.wrapper.get_all_organization_members(organization_name)

        if "user_id" in project_query:
            user_id = project_query["user_id"]
            account_id = self.get_account_id_from_user_id(organization_member_list, user_id)
            if account_id is not None:
                project_query["account_id"] = account_id
            else:
                logger.warning(f"project_query に含まれている user_id: {user_id} のユーザが見つかりませんでした。")

        if "except_user_id" in project_query:
            except_user_id = project_query["except_user_id"]
            except_account_id = self.get_account_id_from_user_id(organization_member_list, except_user_id)
            if except_account_id is not None:
                project_query["except_account_id"] = except_account_id
            else:
                logger.warning(f"project_query に含まれている except_user_id: {except_user_id} のユーザが見つかりませんでした。")

        return project_query

    def get_project_list_from_organization(
        self, organization_name: str, project_query: Optional[Dict[str, Any]] = None
    ) -> List[Project]:
        """
        組織名からプロジェクト一覧を取得する。
        """

        if project_query is not None:
            project_query = self._modify_project_query(organization_name, project_query)

        logger.debug(f"project_query: {project_query}")
        project_list = self.service.wrapper.get_all_projects_of_organization(
            organization_name, query_params=project_query
        )

        for project in project_list:
            project["organization_name"] = organization_name

        return project_list


class ListProject(AbstractCommandLineInterface):
    """
    プロジェクト一覧を表示する。
    """

    def main(self):
        args = self.args
        project_query = {}

        organization_name = args.organization
        main_obj = ListProjectMain(self.service)
        if organization_name is not None:
            if not args.include_not_joined_project:
                # 所属しているプロジェクトのみ出力
                project_query.update({"user_id": self.service.api.login_user_id})

            if args.project_query is not None:
                project_query.update(annofabcli.common.cli.get_json_from_args(args.project_query))
            project_list = main_obj.get_project_list_from_organization(organization_name, project_query)
        else:
            assert args.project_id is not None
            project_id_list = annofabcli.common.cli.get_list_from_args(args.project_id)
            project_list = main_obj.get_project_list_from_project_id(project_id_list)

        logger.info(f"プロジェクト一覧の件数: {len(project_list)}")

        if args.format == FormatArgument.MINIMAL_CSV.value:
            df = create_minimal_dataframe(project_list)
            self.print_csv(df)
        else:
            self.print_according_to_format(project_list)


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    ListProject(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    argument_parser = ArgumentParser(parser)

    query_group = parser.add_mutually_exclusive_group(required=True)

    query_group.add_argument(
        "-p",
        "--project_id",
        type=str,
        nargs="+",
        help="対象プロジェクトのproject_idを指定します。`file://`を先頭に付けると、project_idの一覧が記載されたファイルを指定できます。",
    )

    query_group.add_argument("-org", "--organization", type=str, help="対象の組織名を指定してください。")

    parser.add_argument(
        "-pq",
        "--project_query",
        type=str,
        help="プロジェクトの検索クエリをJSON形式で指定します。'--organization'を指定したときのみ有効なオプションです。"
        "指定しない場合は、組織配下のすべてのプロジェクトを取得します。"
        "`file://`を先頭に付けると、JSON形式のファイルを指定できます。"
        "クエリのフォーマットは、[getProjectsOfOrganization API]"
        "(https://annofab.com/docs/api/#operation/getProjectsOfOrganization)のクエリパラメータと同じです。"
        "さらに追加で、`user_id`, `except_user_id` キーも指定できます。"
        "ただし `page`, `limit`キーは指定できません。",
    )

    parser.add_argument(
        "--include_not_joined_project",
        action="store_true",
        help="'--organization'を指定したときのみ有効なオプションです。指定した場合は、所属していないプロジェクトも出力します。"
        "指定しない場合は、自分が所属しているプロジェクトのみ出力します"
        '（`--project_query \'{"user_id":"my_user_id"}\'`が指定されている状態と同じ）',
    )

    argument_parser.add_format(
        choices=[
            FormatArgument.CSV,
            FormatArgument.MINIMAL_CSV,
            FormatArgument.JSON,
            FormatArgument.PRETTY_JSON,
            FormatArgument.PROJECT_ID_LIST,
        ],
        default=FormatArgument.CSV,
    )
    argument_parser.add_output()
    argument_parser.add_csv_format()

    argument_parser.add_query()
    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "list"
    subcommand_help = "プロジェクト一覧を出力します。"
    description = "プロジェクト一覧を出力します。"

    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description)
    parse_args(parser)



import argparse
import logging
import sys
import uuid
from dataclasses import dataclass
from functools import partial
from multiprocessing import Pool
from pathlib import Path
from typing import Dict, List, Optional

import annofabapi
import pandas
import requests
from annofabapi.models import ProjectMemberRole, SupplementaryData
from dataclasses_json import DataClassJsonMixin
from more_itertools import first_true

import annofabcli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import (
    AbstractCommandLineInterface,
    ArgumentParser,
    build_annofabapi_resource_and_login,
    prompt_yesnoall,
)
from annofabcli.common.utils import get_file_scheme_path

logger = logging.getLogger(__name__)


@dataclass
class CsvSupplementaryData(DataClassJsonMixin):
    """
    CSVに記載されている補助情報
    """

    input_data_id: str
    supplementary_data_number: int
    supplementary_data_name: str
    supplementary_data_path: str
    supplementary_data_id: Optional[str]
    supplementary_data_type: Optional[str]


@dataclass
class SupplementaryDataForPut:
    """
    putする補助情報
    """

    input_data_id: str
    supplementary_data_id: str
    supplementary_data_name: str
    supplementary_data_path: str
    supplementary_data_type: Optional[str]
    supplementary_data_number: int
    last_updated_datetime: Optional[str]


class SubPutSupplementaryData:
    """
    1個の補助情報を登録するためのクラス。multiprocessing.Pool対応。

    Args:
        service:
        facade:
        all_yes:
    """

    def __init__(self, service: annofabapi.Resource, facade: AnnofabApiFacade, all_yes: bool = False):
        self.service = service
        self.facade = facade
        self.all_yes = all_yes
        self.supplementary_data_cache: Dict[str, List[SupplementaryData]] = {}

    def put_supplementary_data(self, project_id: str, supplementary_data: SupplementaryDataForPut):

        file_path = get_file_scheme_path(supplementary_data.supplementary_data_path)
        if file_path is not None:
            request_body = {
                "supplementary_data_name": supplementary_data.supplementary_data_name,
                "supplementary_data_number": supplementary_data.supplementary_data_number,
                "last_updated_datetime": supplementary_data.last_updated_datetime,
            }
            # 省略時は put_supplementary_data_from_file に推定させたいので、Noneも入れない
            if supplementary_data.supplementary_data_type is not None:
                request_body.update({"supplementary_data_type": supplementary_data.supplementary_data_type})

            logger.debug(
                f"'{file_path}'を補助情報として登録します。supplementary_data_name={supplementary_data.supplementary_data_name}"
            )
            self.service.wrapper.put_supplementary_data_from_file(
                project_id,
                input_data_id=supplementary_data.input_data_id,
                supplementary_data_id=supplementary_data.supplementary_data_id,
                file_path=file_path,
                request_body=request_body,
            )

        else:
            supplementary_data_type = supplementary_data.supplementary_data_type
            if supplementary_data_type is None:
                supplementary_data_type = (
                    "text" if supplementary_data.supplementary_data_path.endswith(".txt") else "image"
                )

            request_body = {
                "supplementary_data_name": supplementary_data.supplementary_data_name,
                "supplementary_data_number": supplementary_data.supplementary_data_number,
                "supplementary_data_path": supplementary_data.supplementary_data_path,
                "supplementary_data_type": supplementary_data_type,
                "last_updated_datetime": supplementary_data.last_updated_datetime,
            }

            self.service.api.put_supplementary_data(
                project_id,
                supplementary_data.input_data_id,
                supplementary_data.supplementary_data_id,
                request_body=request_body,
            )

    def confirm_processing(self, confirm_message: str) -> bool:
        """
        `all_yes`属性を見て、処理するかどうかユーザに問い合わせる。
        "ALL"が入力されたら、`all_yes`属性をTrueにする

        Args:
            task_id: 処理するtask_id
            confirm_message: 確認メッセージ

        Returns:
            True: Yes, False: No

        """
        if self.all_yes:
            return True

        yes, all_yes = prompt_yesnoall(confirm_message)

        if all_yes:
            self.all_yes = True

        return yes

    def confirm_put_supplementary_data(
        self, csv_supplementary_data: CsvSupplementaryData, already_exists: bool = False
    ) -> bool:
        message_for_confirm = (
            f"supplementary_data_name='{csv_supplementary_data.supplementary_data_name}' の補助情報を登録しますか？"
        )
        if already_exists:
            message_for_confirm += f"supplementary_data_id={csv_supplementary_data.supplementary_data_id} を上書きします。"
        return self.confirm_processing(message_for_confirm)

    def get_supplementary_data_list_cached(self, project_id: str, input_data_id: str) -> List[SupplementaryData]:
        key = f"{project_id},{input_data_id}"
        if key not in self.supplementary_data_cache:
            supplementary_data_list, _ = self.service.api.get_supplementary_data_list(project_id, input_data_id)
            self.supplementary_data_cache[key] = supplementary_data_list if supplementary_data_list is not None else []
        return self.supplementary_data_cache[key]

    def get_supplementary_data_by_id(
        self, project_id: str, input_data_id: str, supplementary_data_id: str
    ) -> Optional[SupplementaryData]:
        cached_list = self.get_supplementary_data_list_cached(project_id, input_data_id)
        return first_true(cached_list, pred=lambda e: e["supplementary_data_id"] == supplementary_data_id)

    def get_supplementary_data_by_number(
        self, project_id: str, input_data_id: str, supplementary_data_number: int
    ) -> Optional[SupplementaryData]:
        cached_list = self.get_supplementary_data_list_cached(project_id, input_data_id)
        return first_true(cached_list, pred=lambda e: e["supplementary_data_number"] == supplementary_data_number)

    def put_supplementary_data_main(
        self, project_id: str, csv_supplementary_data: CsvSupplementaryData, overwrite: bool = False
    ) -> bool:
        last_updated_datetime = None
        input_data_id = csv_supplementary_data.input_data_id
        supplementary_data_id = csv_supplementary_data.supplementary_data_id
        supplementary_data_path = csv_supplementary_data.supplementary_data_path

        if supplementary_data_id is not None:
            old_supplementary_data_key = f"supplementary_data_id={supplementary_data_id}"
            old_supplementary_data = self.get_supplementary_data_by_id(project_id, input_data_id, supplementary_data_id)
        else:
            supplementary_data_number = csv_supplementary_data.supplementary_data_number
            old_supplementary_data_key = (
                f"input_data_id={input_data_id}, supplementary_data_number={supplementary_data_number}"
            )
            old_supplementary_data = self.get_supplementary_data_by_number(
                project_id, input_data_id, supplementary_data_number
            )
            supplementary_data_id = (
                old_supplementary_data["supplementary_data_id"]
                if old_supplementary_data is not None
                else str(uuid.uuid4())
            )

        if old_supplementary_data is not None:
            if overwrite:
                logger.debug(f"{old_supplementary_data_key} はすでに存在します。")
                last_updated_datetime = old_supplementary_data["updated_datetime"]
            else:
                logger.debug(f"{old_supplementary_data_key} がすでに存在するのでスキップします。")
                return False

        file_path = get_file_scheme_path(supplementary_data_path)
        logger.debug(f"csv_supplementary_data={csv_supplementary_data}")
        if file_path is not None:
            if not Path(file_path).exists():
                logger.warning(f"{supplementary_data_path} は存在しません。")
                return False

        if not self.confirm_put_supplementary_data(
            csv_supplementary_data, already_exists=(last_updated_datetime is not None)
        ):
            return False

        # 補助情報を登録
        supplementary_data_for_put = SupplementaryDataForPut(
            input_data_id=csv_supplementary_data.input_data_id,
            supplementary_data_id=supplementary_data_id,
            supplementary_data_name=csv_supplementary_data.supplementary_data_name,
            supplementary_data_path=csv_supplementary_data.supplementary_data_path,
            supplementary_data_type=csv_supplementary_data.supplementary_data_type,
            supplementary_data_number=csv_supplementary_data.supplementary_data_number,
            last_updated_datetime=last_updated_datetime,
        )
        try:
            self.put_supplementary_data(project_id, supplementary_data_for_put)
            logger.debug(
                f"補助情報を登録しました。"
                f"input_data_id={supplementary_data_for_put.input_data_id},"
                f"supplementary_data_id={supplementary_data_for_put.supplementary_data_id}, "
                f"supplementary_data_name={supplementary_data_for_put.supplementary_data_name}"
            )
            return True

        except requests.exceptions.HTTPError as e:
            logger.warning(e)
            logger.warning(
                f"補助情報の登録に失敗しました。"
                f"input_data_id={supplementary_data_for_put.input_data_id},"
                f"supplementary_data_id={supplementary_data_for_put.supplementary_data_id}, "
                f"supplementary_data_name={supplementary_data_for_put.supplementary_data_name}"
            )
            return False


class PutSupplementaryData(AbstractCommandLineInterface):
    """
    補助情報をCSVで登録する。
    """

    def put_supplementary_data_list(
        self,
        project_id: str,
        supplementary_data_list: List[CsvSupplementaryData],
        overwrite: bool = False,
        parallelism: Optional[int] = None,
    ) -> None:
        """
        補助情報を一括で登録する。

        Args:
            project_id: 補助情報の登録先プロジェクトのプロジェクトID
            supplementary_data_list: 補助情報List
            overwrite: Trueならば、supplementary_data_id（省略時はsupplementary_data_number）がすでに存在していたら上書きします。Falseならばスキップします。
            parallelism: 並列度

        """

        project_title = self.facade.get_project_title(project_id)
        logger.info(f"{project_title} に、{len(supplementary_data_list)} 件の補助情報を登録します。")

        count_put_supplementary_data = 0

        obj = SubPutSupplementaryData(service=self.service, facade=self.facade, all_yes=self.all_yes)
        if parallelism is not None:
            partial_func = partial(obj.put_supplementary_data_main, project_id, overwrite=overwrite)
            with Pool(parallelism) as pool:
                result_bool_list = pool.map(partial_func, supplementary_data_list)
                count_put_supplementary_data = len([e for e in result_bool_list if e])

        else:
            for csv_supplementary_data in supplementary_data_list:
                result = obj.put_supplementary_data_main(
                    project_id, csv_supplementary_data=csv_supplementary_data, overwrite=overwrite
                )
                if result:
                    count_put_supplementary_data += 1

        logger.info(f"{project_title} に、{count_put_supplementary_data} / {len(supplementary_data_list)} 件の補助情報を登録しました。")

    @staticmethod
    def get_supplementary_data_list_from_csv(csv_path: Path) -> List[CsvSupplementaryData]:
        def create_supplementary_data(e):
            supplementary_data_id = e.supplementary_data_id if not pandas.isna(e.supplementary_data_id) else None
            supplementary_data_type = e.supplementary_data_type if not pandas.isna(e.supplementary_data_type) else None
            return CsvSupplementaryData(
                input_data_id=e.input_data_id,
                supplementary_data_number=e.supplementary_data_number,
                supplementary_data_name=e.supplementary_data_name,
                supplementary_data_path=e.supplementary_data_path,
                supplementary_data_id=supplementary_data_id,
                supplementary_data_type=supplementary_data_type,
            )

        df = pandas.read_csv(
            str(csv_path),
            sep=",",
            header=None,
            names=(
                "input_data_id",
                "supplementary_data_number",
                "supplementary_data_name",
                "supplementary_data_path",
                "supplementary_data_id",
                "supplementary_data_type",
            ),
        )
        supplementary_data_list = [create_supplementary_data(e) for e in df.itertuples()]
        return supplementary_data_list

    @staticmethod
    def validate(args: argparse.Namespace) -> bool:
        COMMON_MESSAGE = "annofabcli supplementary_data put: error:"
        if args.csv is not None:
            if not Path(args.csv).exists():
                print(f"{COMMON_MESSAGE} argument --csv: ファイルパスが存在しません。 '{args.csv}'", file=sys.stderr)
                return False

            if args.parallelism is not None and not args.yes:
                print(
                    f"{COMMON_MESSAGE} argument --parallelism: '--parallelism'を指定するときは、必ず'--yes'を指定してください。",
                    file=sys.stderr,
                )
                return False

        return True

    def main(self):
        args = self.args
        if not self.validate(args):
            return

        project_id = args.project_id
        super().validate_project(project_id, [ProjectMemberRole.OWNER])

        supplementary_data_list = self.get_supplementary_data_list_from_csv(Path(args.csv))
        self.put_supplementary_data_list(
            project_id,
            supplementary_data_list=supplementary_data_list,
            overwrite=args.overwrite,
            parallelism=args.parallelism,
        )


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    PutSupplementaryData(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    argument_parser = ArgumentParser(parser)

    argument_parser.add_project_id()

    parser.add_argument(
        "--csv",
        type=str,
        required=True,
        help=(
            "補助情報が記載されたCVファイルのパスを指定してください。"
            "CSVのフォーマットは、「1列目:input_data_id(required), 2列目:supplementary_data_number(required), "
            "3列目:supplementary_data_name(required), 4列目:supplementary_data_path(required), 5列目:supplementary_data_id, "
            "6列目:supplementary_data_type, ヘッダ行なし, カンマ区切り」です。"
            "supplementary_data_pathの先頭が`file://`の場合、ローカルのファイルを補助情報として登録します。 "
            "supplementary_data_idが空の場合はUUIDv4になります。"
            "各項目の詳細は `putSupplementaryData` API を参照してください。"
        ),
    )

    parser.add_argument(
        "--overwrite",
        action="store_true",
        help="指定した場合、supplementary_data_id（省略時はsupplementary_data_number）がすでに存在していたら上書きします。指定しなければ、スキップします。",
    )

    parser.add_argument("--parallelism", type=int, help="並列度。指定しない場合は、逐次的に処理します。必ず'--yes'を指定してください。")

    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "put"
    subcommand_help = "補助情報を登録します。"
    description = "CSVに記載された補助情報を登録します。"
    epilog = "オーナロールを持つユーザで実行してください。"

    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description, epilog=epilog)
    parse_args(parser)



import argparse
import logging
from pathlib import Path
from typing import List, Optional

import pandas

import annofabcli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import AbstractCommandLineInterface, build_annofabapi_resource_and_login
from annofabcli.statistics.linegraph import LineGraph, OutputTarget
from annofabcli.statistics.table import Table

logger = logging.getLogger(__name__)


def write_linegraph_per_user(
    csv: Path, output_dir: Path, user_id_list: Optional[List[str]] = None, minimal_output: bool = False
) -> None:
    """
    折れ線グラフをユーザごとにプロットする。

    Args:
        user_id_list: 折れ線グラフに表示するユーザ

    Returns:

    """
    task_df = pandas.read_csv(str(csv))
    if len(task_df) == 0:
        logger.warning(f"タスク一覧が0件のため、折れ線グラフを出力しません。")
        return

    linegraph_obj = LineGraph(outdir=str(output_dir))

    output_target_list: Optional[List[OutputTarget]] = None
    if minimal_output:
        output_target_list = [OutputTarget.ANNOTATION]

    task_df = Table.create_gradient_df(task_df)
    task_cumulative_df_by_annotator = Table.create_cumulative_df_by_first_annotator(task_df)
    linegraph_obj.write_cumulative_line_graph_for_annotator(
        df=task_cumulative_df_by_annotator,
        output_target_list=output_target_list,
        first_annotation_user_id_list=user_id_list,
    )

    task_cumulative_df_by_inspector = Table.create_cumulative_df_by_first_inspector(task_df)
    linegraph_obj.write_cumulative_line_graph_for_inspector(
        df=task_cumulative_df_by_inspector,
        output_target_list=output_target_list,
        first_inspection_user_id_list=user_id_list,
    )

    task_cumulative_df_by_acceptor = Table.create_cumulative_df_by_first_acceptor(task_df)
    linegraph_obj.write_cumulative_line_graph_for_acceptor(
        df=task_cumulative_df_by_acceptor,
        output_target_list=output_target_list,
        first_acceptance_user_id_list=user_id_list,
    )

    if not minimal_output:
        df_by_date_user_for_annotation = Table.create_dataframe_by_date_user_for_annotation(task_df)
        if len(df_by_date_user_for_annotation) > 0:
            linegraph_obj.write_productivity_line_graph_for_annotator(
                df=df_by_date_user_for_annotation, first_annotation_user_id_list=user_id_list
            )
            linegraph_obj.write_gradient_graph_for_annotator(
                df=task_cumulative_df_by_annotator, first_annotation_user_id_list=user_id_list
            )

        df_by_date_user_for_inspection = Table.create_dataframe_by_date_user_for_inspection(task_df)
        if len(df_by_date_user_for_inspection) > 0:
            linegraph_obj.write_productivity_line_graph_for_inspector(
                df=df_by_date_user_for_inspection, first_inspection_user_id_list=user_id_list
            )
            linegraph_obj.write_gradient_for_inspector(
                df=task_cumulative_df_by_inspector, first_inspection_user_id_list=user_id_list
            )

        df_by_date_user_for_acceptance = Table.create_dataframe_by_date_user_for_acceptance(task_df)
        if len(df_by_date_user_for_acceptance) > 0:
            linegraph_obj.write_productivity_line_graph_for_acceptor(
                df=df_by_date_user_for_acceptance, first_acceptance_user_id_list=user_id_list
            )
            linegraph_obj.write_gradient_for_acceptor(
                df=task_cumulative_df_by_acceptor, first_acceptance_user_id_list=user_id_list
            )


class WriteLingraphPerUser(AbstractCommandLineInterface):
    def main(self):
        args = self.args
        user_id_list = annofabcli.common.cli.get_list_from_args(args.user_id) if args.user_id is not None else None
        write_linegraph_per_user(
            csv=args.csv, output_dir=args.output_dir, user_id_list=user_id_list, minimal_output=args.minimal
        )


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    WriteLingraphPerUser(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--csv",
        type=Path,
        required=True,
        help=("CSVファイルのパスを指定してください。" "CSVは、'statistics visualize'コマンドの出力結果である'タスクlist.csv'と同じフォーマットです。"),
    )

    parser.add_argument(
        "-u",
        "--user_id",
        nargs="+",
        help=(
            "絞り込み条件となるユーザのuser_idを指定してください。" "指定しない場合は、上位20人のユーザ情報がプロットされます。" "file://`を先頭に付けると、一覧が記載されたファイルを指定できます。"
        ),
    )

    parser.add_argument(
        "--minimal",
        action="store_true",
        help="必要最小限のファイルを出力します。",
    )

    parser.add_argument("-o", "--output_dir", type=Path, required=True, help="出力ディレクトリのパス")

    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "write_linegraph_per_user"
    subcommand_help = "CSVからユーザごとの指標を折れ線グラフで出力します。"
    description = "CSVからユーザごとの指標を折れ線グラフで出力します。"
    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description)
    parse_args(parser)



import argparse
import json
import logging
from collections import defaultdict
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import annofabapi
import pandas
from annofabapi.models import ProjectMemberRole, TaskHistory, TaskPhase

import annofabcli
import annofabcli.common.cli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import AbstractCommandLineInterface, ArgumentParser, build_annofabapi_resource_and_login
from annofabcli.common.download import DownloadingFile
from annofabcli.common.utils import isoduration_to_hour

logger = logging.getLogger(__name__)


class ListSubmittedTaskCountMain:
    def __init__(self, service: annofabapi.Resource):
        self.service = service
        self.facade = AnnofabApiFacade(service)

    @staticmethod
    def from_datetime_to_date(datetime: str) -> str:
        return datetime[0:10]

    @staticmethod
    def _get_actual_worktime_hour(labor: Dict[str, Any]) -> float:
        working_time_by_user = labor["values"]["working_time_by_user"]
        if working_time_by_user is None:
            return 0

        value = working_time_by_user.get("results")
        if value is None:
            return 0
        else:
            return value / 3600 / 1000

    @staticmethod
    def to_formatted_dataframe(
        submitted_task_count_df: pandas.DataFrame,
        account_statistics_df: pandas.DataFrame,
        labor_df: pandas.DataFrame,
        user_df: pandas.DataFrame,
    ):
        df = (
            submitted_task_count_df.merge(account_statistics_df, how="outer", on=["date", "account_id"])
            .merge(labor_df, how="outer", on=["date", "account_id"])
            .fillna(0)
            .merge(user_df, how="inner", on="account_id")
        )
        df.sort_values(["date", "user_id"], inplace=True)
        columns = [
            "date",
            "user_id",
            "username",
            "biography",
            "monitored_worktime_hour",
            "actual_worktime_hour",
            "annotation_submitted_task_count",
            "inspection_submitted_task_count",
            "acceptance_submitted_task_count",
            "rejected_task_count",
        ]

        return df[columns]

    def get_task_history_dict(
        self, project_id: str, task_history_json_path: Optional[Path]
    ) -> Dict[str, List[TaskHistory]]:
        if task_history_json_path is not None:
            json_path = task_history_json_path
        else:
            cache_dir = annofabcli.utils.get_cache_dir()
            json_path = cache_dir / f"task-history-{project_id}.json"
            downloading_obj = DownloadingFile(self.service)
            downloading_obj.download_task_history_json(project_id, dest_path=str(json_path))

        logger.debug(f"タスク履歴全件ファイルを読み込み中。{json_path}")
        with json_path.open(encoding="utf-8") as f:
            task_history_dict = json.load(f)
            return task_history_dict

    def create_user_df(self, project_id: str):
        member_list = self.facade.get_organization_members_from_project_id(project_id)
        return pandas.DataFrame(member_list, columns=["account_id", "user_id", "username", "biography"])

    @staticmethod
    def _is_contained_daterange(target_date: str, start_date: Optional[str] = None, end_date: Optional[str] = None):
        if start_date is not None:
            if target_date < start_date:
                return False
        if end_date is not None:
            if target_date > end_date:
                return False
        return True

    def create_account_statistics_df(
        self, project_id: str, start_date: Optional[str] = None, end_date: Optional[str] = None
    ) -> pandas.DataFrame:
        account_statistics = self.service.wrapper.get_account_statistics(project_id)
        data_list: List[Dict[str, Any]] = []
        for stat_by_user in account_statistics:
            account_id = stat_by_user["account_id"]
            histories = stat_by_user["histories"]
            for stat in histories:
                data = {
                    "account_id": account_id,
                    "monitored_worktime_hour": isoduration_to_hour(stat["worktime"]),
                    "rejected_task_count": stat["tasks_rejected"],
                    "date": stat["date"],
                }
                if not self._is_contained_daterange(data["date"], start_date=start_date, end_date=end_date):
                    continue

                if data["monitored_worktime_hour"] == 0 and data["rejected_task_count"] == 0:
                    continue

                data_list.append(data)

        if len(data_list) > 0:
            return pandas.DataFrame(data_list)
        else:
            return pandas.DataFrame(columns=["date", "account_id", "monitored_worktime_hour", "rejected_task_count"])

    def create_labor_df(
        self, project_id: str, start_date: Optional[str] = None, end_date: Optional[str] = None
    ) -> pandas.DataFrame:
        def to_new_labor(e: Dict[str, Any]) -> Dict[str, Any]:
            return dict(
                date=e["date"],
                account_id=e["account_id"],
                actual_worktime_hour=self._get_actual_worktime_hour(e),
            )

        labor_list: List[Dict[str, Any]] = self.service.api.get_labor_control(
            {"project_id": project_id, "from": start_date, "to": end_date}
        )[0]
        new_labor_list = [
            to_new_labor(e) for e in labor_list if e["account_id"] is not None and self._get_actual_worktime_hour(e) > 0
        ]
        if len(new_labor_list) > 0:
            return pandas.DataFrame(new_labor_list)
        else:
            return pandas.DataFrame(columns=["date", "account_id", "actual_worktime_hour"])

    def create_submitted_task_count_df(
        self,
        task_history_dict: Dict[str, List[TaskHistory]],
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
    ) -> pandas.DataFrame:
        def _set_zero_if_not_exists(df: pandas.DataFrame):
            for phase in TaskPhase:
                col = f"{phase.value}_submitted_task_count"
                if col not in df.columns:
                    df[col] = 0

        task_history_count_dict: Dict[Tuple[str, str, str], int] = defaultdict(int)
        for _, task_history_list in task_history_dict.items():
            for task_history in task_history_list:
                if task_history["ended_datetime"] is not None and task_history["account_id"] is not None:
                    ended_date = self.from_datetime_to_date(task_history["ended_datetime"])
                    task_history_count_dict[(task_history["account_id"], task_history["phase"], ended_date)] += 1

        data_list = []
        for key, task_count in task_history_count_dict.items():
            account_id, phaes, date = key
            data: Dict[str, Any] = {"date": date, "phase": phaes, "account_id": account_id, "task_count": task_count}
            if not self._is_contained_daterange(data["date"], start_date=start_date, end_date=end_date):
                continue

            data_list.append(data)

        if len(data_list) == 0:
            return pandas.DataFrame(
                columns=[
                    "date",
                    "account_id",
                    "annotation_submitted_task_count",
                    "inspection_submitted_task_count",
                    "acceptance_submitted_task_count",
                ]
            )

        df = pandas.DataFrame(data_list)
        df2 = df.pivot_table(columns="phase", values="task_count", index=["date", "account_id"]).fillna(0)
        df2.rename(
            columns={
                "annotation": "annotation_submitted_task_count",
                "inspection": "inspection_submitted_task_count",
                "acceptance": "acceptance_submitted_task_count",
            },
            inplace=True,
        )
        _set_zero_if_not_exists(df2)
        return df2

    def create_user_statistics_by_date(
        self,
        project_id: str,
        task_history_json_path: Optional[Path],
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
    ) -> pandas.DataFrame:
        task_history_dict = self.get_task_history_dict(project_id, task_history_json_path)

        submitted_task_count_df = self.create_submitted_task_count_df(
            task_history_dict=task_history_dict, start_date=start_date, end_date=end_date
        )
        labor_df = self.create_labor_df(project_id, start_date=start_date, end_date=end_date)
        account_statistics_df = self.create_account_statistics_df(project_id, start_date=start_date, end_date=end_date)
        user_df = self.create_user_df(project_id)

        df2 = self.to_formatted_dataframe(submitted_task_count_df, account_statistics_df, labor_df, user_df)
        return df2


class ListSubmittedTaskCountArgs(AbstractCommandLineInterface):
    def main(self):
        args = self.args

        project_id = args.project_id
        super().validate_project(
            project_id, project_member_roles=[ProjectMemberRole.OWNER, ProjectMemberRole.TRAINING_DATA_USER]
        )

        main_obj = ListSubmittedTaskCountMain(service=self.service)
        df = main_obj.create_user_statistics_by_date(
            project_id, args.task_history_json, start_date=args.start_date, end_date=args.end_date
        )
        self.print_csv(df)


def parse_args(parser: argparse.ArgumentParser):
    argument_parser = ArgumentParser(parser)

    argument_parser.add_project_id()

    parser.add_argument(
        "--task_history_json",
        type=Path,
        help="タスク履歴情報が記載されたJSONファイルのパスを指定してます。JSONファイルは`$ annofabcli project download task_history`コマンドで取得できます。"
        "指定しない場合は、AnnoFabからタスク履歴全件ファイルをダウンロードします。",
    )

    parser.add_argument("--start_date", type=str, help="集計対象の開始日(YYYY-mm-dd)")
    parser.add_argument("--end_date", type=str, help="集計対象の終了日(YYYY-mm-dd)")

    argument_parser.add_csv_format()
    argument_parser.add_output()

    parser.set_defaults(subcommand_func=main)


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    ListSubmittedTaskCountArgs(service, facade, args).main()


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "list_by_date_user"
    subcommand_help = "タスク数や作業時間などの情報を、日ごとユーザごとに出力します。"
    description = "タスク数や作業時間などの情報を、日ごとユーザごとに出力します。"
    epilog = "オーナロールまたはアノテーションユーザロールを持つユーザで実行してください。"
    parser = annofabcli.common.cli.add_parser(
        subparsers, subcommand_name, subcommand_help, description=description, epilog=epilog
    )
    parse_args(parser)



import argparse
import logging
from typing import List

import annofabapi
from annofabapi.models import JobType, ProjectMemberRole

import annofabcli
from annofabcli import AnnofabApiFacade
from annofabcli.common.cli import (
    AbstractCommandLineInterface,
    ArgumentParser,
    build_annofabapi_resource_and_login,
    get_list_from_args,
)

logger = logging.getLogger(__name__)


class DeleteJobMain:
    def __init__(self, service: annofabapi.Resource):
        self.service = service
        self.facade = AnnofabApiFacade(service)

    def delete_job_list(self, project_id: str, job_type: JobType, job_id_list: List[str]):
        for job_id in job_id_list:
            logger.debug(f"job_id={job_id} のジョブを削除します。")
            try:
                self.service.api.delete_project_job(project_id, job_type.value, job_id)
            except Exception as e:  # pylint: disable=broad-except
                logger.warning(e)


class DeleteJob(AbstractCommandLineInterface):
    def main(self):
        args = self.args
        project_id = args.project_id
        super().validate_project(project_id, project_member_roles=[ProjectMemberRole.OWNER])

        job_type = JobType(args.job_type)
        job_id_list = get_list_from_args(args.job_id)

        main_obj = DeleteJobMain(self.service)
        main_obj.delete_job_list(args.project_id, job_type=job_type, job_id_list=job_id_list)


def main(args):
    service = build_annofabapi_resource_and_login(args)
    facade = AnnofabApiFacade(service)
    DeleteJob(service, facade, args).main()


def parse_args(parser: argparse.ArgumentParser):
    argument_parser = ArgumentParser(parser)

    job_choices = [e.value for e in JobType]
    argument_parser.add_project_id()

    parser.add_argument("--job_type", type=str, choices=job_choices, required=True, help="ジョブタイプを指定します。")
    parser.add_argument(
        "--job_id",
        type=str,
        nargs="+",
        required=True,
        help="削除するジョブのjob_idを指定します。" + "`file://`を先頭に付けると、job_idの一覧が記載されたファイルを指定できます。",
    )

    parser.set_defaults(subcommand_func=main)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "delete"
    subcommand_help = "ジョブを削除する。"
    description = "ジョブを削除する。"
    epilog = "オーナロールで実行してください。"
    parser = annofabcli.common.cli.add_parser(subparsers, subcommand_name, subcommand_help, description, epilog=epilog)
    parse_args(parser)



import argparse

import annofabcli
import annofabcli.common.cli
import annofabcli.statistics.list_annotation_count
import annofabcli.statistics.list_by_date_user
import annofabcli.statistics.list_cumulative_labor_time
import annofabcli.statistics.list_labor_time_per_user
import annofabcli.statistics.list_task_progress
import annofabcli.statistics.merge_visualization_dir
import annofabcli.statistics.summarize_task_count
import annofabcli.statistics.summarize_task_count_by_task_id
import annofabcli.statistics.summarize_task_count_by_user
import annofabcli.statistics.visualize_statistics


def parse_args(parser: argparse.ArgumentParser):

    subparsers = parser.add_subparsers(dest="subcommand_name")

    # サブコマンドの定義
    annofabcli.statistics.list_annotation_count.add_parser(subparsers)
    annofabcli.statistics.list_cumulative_labor_time.add_parser(subparsers)
    annofabcli.statistics.list_labor_time_per_user.add_parser(subparsers)
    annofabcli.statistics.list_by_date_user.add_parser(subparsers)
    annofabcli.statistics.list_task_progress.add_parser(subparsers)
    annofabcli.statistics.merge_visualization_dir.add_parser(subparsers)
    annofabcli.statistics.summarize_task_count.add_parser(subparsers)
    annofabcli.statistics.summarize_task_count_by_task_id.add_parser(subparsers)
    annofabcli.statistics.summarize_task_count_by_user.add_parser(subparsers)
    annofabcli.statistics.visualize_statistics.add_parser(subparsers)


def add_parser(subparsers: argparse._SubParsersAction):
    subcommand_name = "statistics"
    subcommand_help = "統計関係のサブコマンド"
    description = "統計関係のサブコマンド"

    parser = annofabcli.common.cli.add_parser(
        subparsers, subcommand_name, subcommand_help, description, is_subcommand=False
    )
    parse_args(parser)



"""
型タイプを定義する
"""
from typing import Any, Dict

# 型タイプ
Task = Dict[str, Any]
Inspection = Dict[str, Any]
Annotation = Dict[str, Any]
TaskHistory = Dict[str, Any]
TaskHistoryEvent = Dict[str, Any]
TaskId = str
InputDataId = str
